<?php

/**
* @file 
* CSS Sprite Generator
* @author Tj Holowaychuk <http://www.350designs.com/>
* @package Sprites
*/
 
/**
* Initialize sprite generation processes.
*/
function sprites_generate() {
  global $sprites_created;
  $per_image  = variable_get('sprites_per_image',  SPRITES_PER_IMAGE);
  $max_width  = variable_get('sprites_max_width',  SPRITES_MAX_WIDTH);               
  $max_height = variable_get('sprites_max_height', SPRITES_MAX_HEIGHT);     
  $offsetx    = variable_get('sprites_offsetx',    SPRITES_OFFSET_X);         
  $offsety    = variable_get('sprites_offsety',    SPRITES_OFFSET_Y);         
  $formats    = array('gif', 'png', 'jpg');    
  
  sprites_collection_delete_all();
  
  foreach($formats AS $format){
    $images = sprites_images_get_filepaths($format);           
    
    if (!count($images)){
      continue;
    }          
              
    // Validate images
    sprites_images_validate($images, $max_width, $max_height);
               
    // Sprites per collection limit
    if ($per_image){         
      $chunks = array_chunk($images, $per_image);  
      foreach($chunks AS $i => $chunk){           
        $name  = 'sprites_' . ($i + 1) . '.' . $format; 
        $rules = _sprites_create($chunk, $max_width, $max_height, $name, 'sprites', FALSE, $offsetx, $offsety); 
        _sprites_generate($name, $rules, $set_css); 
      }
    }
    else {
      $rules = _sprites_create($images, $max_width, $max_height, 'sprites.' . $format, 'sprites', FALSE, $offsetx, $offsety);  
      _sprites_generate('sprites.' . $format, $rules, $set_css);    
    }  
  }  
  
  // Tally up sprites created
  if (count($sprites_created)){                                       
    foreach($sprites_created AS $format => $count){
      drupal_set_message(t('!count !format sprites generated.', array('!count' => $count, '!format' => $format)));
    }
  }
  
  // Rewrite css
  sprites_css_rewrite();                          
}                               

/**
* Helper for sprites_generate();
* 
* @see sprites_generate
*/
function _sprites_generate($name, $rules, $set_css) {  
  if (count($rules)){       
    // Save sprites object
    $collection = new stdClass();
    $collection->sid      = 0;
    $collection->filename = file_directory_path() . '/sprites/' . $name; 
    $collection->rules    = $rules; 
    sprites_collection_save($collection);
  }
}  
     
/**
* Helper for sprites_create();
* 
* @todo: refactor large mess which makes up the sprite object's
* @todo: play around with image compression and test on a variety of themes
* @see sprites_create
*/  
function _sprites_create($images, $max_width = SPRITES_MAX_WIDTH, $max_height = SPRITES_MAX_HEIGHT, $name = 'sprites.jpg', $dest = 'sprites', $validate_images = FALSE, $offsetx = SPRITES_OFFSET_X, $offsety = SPRITES_OFFSET_Y){ 
  global $sprites_created;
  $rules    = array(); 
  $filename = $dest ? file_directory_path() . '/' . $dest . '/' . $name : file_directory_path() . '/' . $name; 
  $ext      = pathinfo($filename, PATHINFO_EXTENSION); 
  $_offsety = 0;
  
  // Due to not 'packing' sprites there is currently no
  // use for implementing offsetx.
  $offsetx = 0;
  
  // Destination
  if ($dest = file_create_path($dest)){
    if (!file_check_directory($dest, FILE_CREATE_DIRECTORY)){
      return FALSE;
    }
  }
  else {
    return FALSE;
  }
  
  // Remove images that do not meet the criteria   
  if ($validate_images === TRUE){
    sprites_images_validate($images, $max_width, $max_height);
  }
  
  if (count($images)){
    // Generate CSS rules and offsets 
    foreach($images AS $image){ 
      $rules[$image->filename] = array(
          'image'      => $image,
          'selector'   => '.sprite-' . preg_replace('/[^a-zA-Z0-9]/is', '-', $image->name),
          'background' => $filename, 
          'width'      => $image->info[0],
          'height'     => $image->info[1],
          'offsetx'    => $offsetx,
          'offsety'    => $_offsety,
        );                 
      $_offsety += $image->info[1] + $offsety;  
    }

    // Generate sprites Image
    if (count($rules)){    
      // Find the maximum width used
      foreach($rules AS $rule){
        $widths[] = $rule['width'];
      }                   
      
      // Create image resource     
      if ($ext == 'gif'){ 
        $sprites = imagecreate(max($widths) + $offsetx, $_offsety);  
      }        
      else {
        $sprites = imagecreatetruecolor(max($widths) + $offsetx, $_offsety);
        imagealphablending($sprites, FALSE);
      }
      
      // Image formats which support transparency should have an
      // irregular background color which is later considered transparent
      switch($ext){
        case 'gif':
        case 'png':
          $background_color = imagecolorallocate($sprites, 255, 0, 255);
          break;
        
        case 'jpg':
        case 'jpeg':
          $background_color = imagecolorallocate($sprites, 255, 255, 255); 
          break;
      }
      
      foreach($rules AS $rule){
        if (!file_exists($rule['image']->filename)){
          continue;
        }
              
        // Create source image resource
        switch($rule['image']->info['mime']){
          case 'image/jpg':
          case 'image/jpeg':
            $image = imagecreatefromjpeg($rule['image']->filename); 
            break;
            
          case 'image/png':
            $image = imagecreatefrompng($rule['image']->filename);
            break;
            
          case 'image/gif':
            $image = imagecreatefromgif($rule['image']->filename);
            break;
        }
                   
        // Copy source image to the offset position
        if (!imagecopy($sprites, $image, 0, $rule['offsety'], 0, 0, $rule['width'], $rule['height'])){
          drupal_set_message(t('Error copying image %filename.', array('%filename' => $rule['image']->filename)), 'error');    
        }
        
        @imagedestroy($image);
      }
      
      // Remove existing files
      file_delete($filename);
      
      // Create the appropriate image type based on the extension passed 
      switch($ext){
        case 'jpg':
        case 'jpeg':
          imagejpeg($sprites, $filename, 90); // @todo settings and test 
          break;
          
        case 'png':
          imagesavealpha($sprites, TRUE);  
          imagepng($sprites, $filename, 0); // @todo settings and test
          break;
          
        case 'gif':
          imagecolortransparent($sprites, $background_color);
          imagegif($sprites, $filename); // @todo settings and test
          break;
      } 
      
      @imagedestroy($sprites); 
    }
  }
  
  $sprites_created[$ext] += count($rules);
  
  return $rules;                                
}

/**
* Ignore images that do not follow specific criteria.
*   
* Criteria:
*   - Within width and height limits.
* 
*   - Invoke hook_sprites_images_ignore();  
*     Array of image filenames ex: ('./misc/farbtastic/marker.png') 
*     which will be ignored and not included or re-written as CSS sprites.
* 
*   - Images should be ignored if they are already being used  
*     as sprites, animated gifs, or are simply rarely seen.
* 
* @param array $images    
*   Image objects.
* 
* @param int $max_width
* 
* @param int $max_height
*/
function sprites_images_validate(&$images, $max_width = SPRITES_MAX_WIDTH_DEFAULT, $max_height = SPRITES_MAX_HEIGHT_DEFAULT) { 
  if (count($images)){
    $ignore = module_invoke_all('sprites_images_ignore');
    foreach($images AS $i => $image){
      $image->info = @getimagesize($image->filename);
      if ($image->info[0] > $max_width || $image->info[1] > $max_height || in_array($image->filename, $ignore)){ 
        unset($images[$i]);  
      }
    }
  }   
}   

/**
* Get white list directories.
* 
* @return mixed
*   - success: Array of directories.
*   - failure: FALSE;
*/
function sprites_get_whitelist() {  
  $dirs = @explode("\n", trim((string) variable_get('sprites_whitelist', SPRITES_WHITELIST)));
                      
  if (count($dirs) && $dirs[0]){  
    return $dirs;     
  } 
  else {
    return FALSE;
  }
}

/**
* Get image files from directories that may contain them.
* 
* @param string $format
*   File extension such as 'gif', 'png', etc. 
* 
* @return array
*   Files.
* 
* @todo: test some more
*/
function sprites_images_get_filepaths($format) {
  $files = array();
  
  if ($dirs = sprites_get_whitelist()){
    foreach($dirs AS $dir){  
      if (strpos($dir, '/') === 0){      
        $files = @array_merge(file_scan_directory('.' . $dir, '\.(' .$format. ')$'), $files);  
      }
    }  
  }
      
  return $files;
}

/**
* Rewrite aggregated css to implement the sprites.
* 
* @todo: refactor/clean up
* @see sprites_css_rewrite
*/
function sprites_css_rewrite() {
  $status = array();
  $files = file_scan_directory('./' . file_directory_path() . '/css', '\.(css)$');
  
  if (!count($files)){
    drupal_set_message(t('Failed to rewrite aggregated CSS. Visit !link to enable CSS aggregation.', array('!link' => l('admin/settings/performance', 'admin/settings/performance'))), 'error');
    return FALSE;
  }
  
  // Load sprite collections
  $sprite_collections = sprites_collection_load_all();
  
  if (!count($sprite_collections)){
    drupal_set_message(t('Failed to load sprite collections.'), 'error');
    return FALSE;
  }
  
  // Rewrite each css file                             
  foreach($files AS $filepath => $file){
    $contents = file_get_contents($filepath);
    $contents = _sprites_css_rewrite($contents, $sprite_collections);
    
    if (file_save_data($contents, $filepath, FILE_EXISTS_REPLACE)){
      $status['rewritten']++;    
    }
    else {
      drupal_set_message(t('Failed to rewrite %filepath.', array('%filepath' => $filepath)), 'error');
    }
  }
  
  // Report on what has been accomplished by this subroutine 
  if (count($status)){
    drupal_set_message(t('!count CSS files rewritten to implement CSS sprites.', array('!count' => $status['rewritten'])));
  }
}

/**
* Return css rewritten to implement sprites.
* 
* @see sprites_css_rewrite
*/
function _sprites_css_rewrite($contents, $sprite_collections) {
  // Strip basepath from uris
  if (count($sprite_collections)){
    foreach($sprite_collections AS $cid => $sprite_collection){
      if (count($sprite_collection->rules)){
        foreach($sprite_collection->rules AS $rule){
          $contents = preg_replace('/url\((.*?)' . $rule['image']->basename . '\)/i', 'url(' + $rule['background'] + ')', $contents);      
          // @todo: finish and add background-position offsets
        }
      }
    }
  }  

  return $contents;
}


